'use strict';
Object.defineProperty(exports, '__esModule', { value: true });
exports.insertNativeElementByIndex =
  exports.setTrackUpdate =
  exports.finishCommit =
  exports.commit =
  exports.createNativeElement =
    void 0;
const core_1 = require('@dark-engine/core');
const utils_1 = require('../utils');
const events_1 = require('../events');
const portal_1 = require('../portal');
const constants_1 = require('../constants');
let moves = [];
let patches = [];
let trackUpdate = null;
const createNativeElementMap = {
  [core_1.NodeType.TAG]: vNode => {
    const tagNode = vNode;
    const name = tagNode.name;
    return (0, utils_1.detectIsSvgElement)(name)
      ? document.createElementNS('http://www.w3.org/2000/svg', name)
      : document.createElement(name);
  },
  [core_1.NodeType.TEXT]: vNode => {
    return document.createTextNode(vNode.value);
  },
  [core_1.NodeType.COMMENT]: vNode => {
    return document.createComment(vNode.value);
  },
};
function createNativeElement(node) {
  return createNativeElementMap[node.type](node);
}
exports.createNativeElement = createNativeElement;
function setObjectStyle(element, style) {
  const keys = Object.keys(style);
  for (const key of keys) {
    element.style.setProperty(key, String(style[key]));
  }
}
function addAttributes(element, node, isHydrateZone) {
  const attrNames = Object.keys(node.attrs);
  const tagElement = element;
  for (let attrName of attrNames) {
    const attrValue = node.attrs[attrName];
    const attribute = performAttribute(tagElement, attrName, attrValue);
    if (attribute === null) {
      continue;
    } else {
      attrName = attribute;
    }
    if ((0, events_1.detectIsEvent)(attrName)) {
      (0, events_1.delegateEvent)(tagElement, (0, events_1.getEventName)(attrName), attrValue);
    } else if (!isHydrateZone && !(0, core_1.detectIsUndefined)(attrValue) && !core_1.ATTR_BLACK_LIST[attrName]) {
      const stop = patchProperties({
        tagName: node.name,
        element: tagElement,
        attrValue,
        attrName,
      });
      !stop && tagElement.setAttribute(attrName, attrValue);
    }
  }
}
function updateAttributes(element, prevNode, nextNode) {
  const attrNames = getAttributeNames(prevNode, nextNode);
  const tagElement = element;
  for (let attrName of attrNames) {
    const prevAttrValue = prevNode.attrs[attrName];
    const nextAttrValue = nextNode.attrs[attrName];
    const attribute = performAttribute(tagElement, attrName, nextAttrValue, prevAttrValue);
    if (attribute === null) {
      continue;
    } else {
      attrName = attribute;
    }
    if (!(0, core_1.detectIsUndefined)(nextAttrValue)) {
      if ((0, events_1.detectIsEvent)(attrName)) {
        prevAttrValue !== nextAttrValue &&
          (0, events_1.delegateEvent)(tagElement, (0, events_1.getEventName)(attrName), nextAttrValue);
      } else if (!core_1.ATTR_BLACK_LIST[attrName] && prevAttrValue !== nextAttrValue) {
        const stop = patchProperties({
          tagName: nextNode.name,
          element: tagElement,
          attrValue: nextAttrValue,
          attrName,
        });
        !stop && tagElement.setAttribute(attrName, nextAttrValue);
      }
    } else {
      tagElement.removeAttribute(attrName);
    }
  }
}
function performAttribute(tagElement, attrName, nextAttrValue, prevAttrValue) {
  if (attrName[0] === constants_1.EXCLUDE_ATTR_MARK) return null;
  if (attrName === core_1.REF_ATTR) {
    (0, core_1.applyRef)(nextAttrValue, tagElement);
    return null;
  }
  if (
    (attrName === constants_1.CLASS_ATTR || attrName === constants_1.CLASS_NAME_ATTR) &&
    nextAttrValue !== prevAttrValue
  ) {
    toggleAttribute(tagElement, constants_1.CLASS_ATTR, nextAttrValue);
    return null;
  }
  if (
    attrName === constants_1.STYLE_ATTR &&
    nextAttrValue &&
    nextAttrValue !== prevAttrValue &&
    (0, core_1.detectIsObject)(nextAttrValue)
  ) {
    setObjectStyle(tagElement, nextAttrValue);
    return null;
  }
  if (attrName === constants_1.AS_ATTR) {
    attrName = attrName.slice(1, constants_1.AS_ATTR.length);
  }
  return attrName;
}
function toggleAttribute(element, name, value) {
  value ? element.setAttribute(name, value) : element.removeAttribute(name);
}
function getAttributeNames(prevNode, nextNode) {
  const attrNames = new Set();
  const prevAttrs = Object.keys(prevNode.attrs);
  const nextAttrs = Object.keys(nextNode.attrs);
  const size = Math.max(prevAttrs.length, nextAttrs.length);
  for (let i = 0; i < size; i++) {
    attrNames.add(prevAttrs[i] || nextAttrs[i]);
  }
  return attrNames;
}
function patchProperties(options) {
  const { tagName, element, attrName, attrValue } = options;
  const fn = specialCasesMap[tagName];
  let stop = fn ? fn(element, attrName, attrValue) : false;
  if (canSetProperty(element, attrName)) {
    element[attrName] = attrValue;
  }
  if (!stop && (0, core_1.detectIsBoolean)(attrValue)) {
    stop = !attrName.includes('-');
  }
  return stop;
}
function canSetProperty(element, key) {
  const prototype = Object.getPrototypeOf(element);
  const descriptor = Object.getOwnPropertyDescriptor(prototype, key);
  return Boolean(descriptor?.set);
}
const specialCasesMap = {
  [constants_1.INPUT_TAG]: (element, attrName, attrValue) => {
    if (attrName === constants_1.VALUE_ATTR) {
      patches.push(() => {
        (0, core_1.detectIsBoolean)(attrValue) ? (element.checked = attrValue) : (element.value = String(attrValue));
      });
    }
    return false;
  },
  [constants_1.TEXTAREA_TAG]: (element, attrName, attrValue) => {
    if (attrName === constants_1.VALUE_ATTR) {
      element.innerText = String(attrValue);
      return true;
    }
    return false;
  },
};
function commitCreation(fiber) {
  const parentFiber = (0, core_1.getFiberWithElement)(fiber.parent);
  const parentElement = parentFiber.element;
  const childNodes = parentElement.childNodes;
  const isHydrateZone = (0, core_1.$$scope)().getIsHydrateZone();
  if (isHydrateZone) {
    const nativeElement = childNodes[fiber.eidx];
    if (
      (0, core_1.detectIsTextVirtualNode)(fiber.inst) &&
      nativeElement instanceof Text &&
      fiber.inst.value.length !== nativeElement.length
    ) {
      nativeElement.splitText(fiber.inst.value.length);
    }
    fiber.element = nativeElement;
  } else {
    if (!(fiber.mask & core_1.SHADOW_MASK)) {
      if (childNodes.length === 0 || fiber.eidx > childNodes.length - 1) {
        !(0, utils_1.detectIsVoidElement)(parentFiber.inst.name) && appendNativeElement(fiber.element, parentElement);
      } else {
        insertNativeElement(fiber.element, parentElement.childNodes[fiber.eidx], parentElement);
      }
    }
  }
  (0, core_1.detectIsTagVirtualNode)(fiber.inst) && addAttributes(fiber.element, fiber.inst, isHydrateZone);
}
function commitUpdate(fiber) {
  const element = fiber.element;
  const prevInstance = fiber.alt.inst;
  const nextInstance = fiber.inst;
  (0, core_1.detectIsPlainVirtualNode)(nextInstance)
    ? prevInstance.value !== nextInstance.value && (element.textContent = nextInstance.value)
    : updateAttributes(element, prevInstance, nextInstance);
}
function commitDeletion(fiber) {
  const parentFiber = (0, core_1.getFiberWithElement)(fiber.parent);
  if (fiber.mask & core_1.FLUSH_MASK) {
    parentFiber.element.innerHTML && (parentFiber.element.innerHTML = '');
    return;
  }
  (0, core_1.walk)(fiber, (fiber, skip) => {
    if (fiber.element) {
      !(0, portal_1.detectIsPortal)(fiber.inst) &&
        canRemoveNativeElement(fiber.element, parentFiber.element) &&
        removeNativeElement(fiber.element, parentFiber.element);
      return skip();
    }
  });
}
function move(fiber) {
  const sourceNodes = (0, core_1.collectElements)(fiber, x => x.element);
  const sourceNode = sourceNodes[0];
  const parentElement = sourceNode.parentElement;
  const sourceFragment = new DocumentFragment();
  const elementIdx = fiber.eidx;
  let idx = 0;
  const move = () => {
    for (let i = 1; i < sourceNodes.length; i++) {
      removeNativeElement(parentElement.childNodes[elementIdx + 1], parentElement);
    }
    replaceNativeElement(sourceFragment, parentElement.childNodes[elementIdx], parentElement);
  };
  for (const node of sourceNodes) {
    insertNativeElement(document.createComment(`${elementIdx}:${idx}`), node, parentElement);
    appendNativeElement(node, sourceFragment);
    idx++;
  }
  moves.push(move);
}
const commitMap = {
  [core_1.CREATE_EFFECT_TAG]: fiber => {
    if (!fiber.element || (0, portal_1.detectIsPortal)(fiber.inst)) return;
    trackUpdate && trackUpdate(fiber.element);
    commitCreation(fiber);
  },
  [core_1.UPDATE_EFFECT_TAG]: fiber => {
    fiber.mask & core_1.MOVE_MASK && (move(fiber), (fiber.mask &= ~core_1.MOVE_MASK));
    if (!fiber.element || (0, portal_1.detectIsPortal)(fiber.inst)) return;
    trackUpdate && trackUpdate(fiber.element);
    commitUpdate(fiber);
  },
  [core_1.DELETE_EFFECT_TAG]: commitDeletion,
  [core_1.SKIP_EFFECT_TAG]: core_1.dummyFn,
};
function finishCommit() {
  moves.forEach(x => x());
  patches.forEach(x => x());
  moves = [];
  patches = [];
}
exports.finishCommit = finishCommit;
const commit = fiber => commitMap[fiber.tag](fiber);
exports.commit = commit;
const setTrackUpdate = fn => (trackUpdate = fn);
exports.setTrackUpdate = setTrackUpdate;
const appendNativeElement = (element, parent) => parent.appendChild(element);
const insertNativeElement = (element, sibling, parent) => parent.insertBefore(element, sibling);
const insertNativeElementByIndex = (element, idx, parent) => parent.insertBefore(element, parent.childNodes[idx]);
exports.insertNativeElementByIndex = insertNativeElementByIndex;
const replaceNativeElement = (element, candidate, parent) => parent.replaceChild(element, candidate);
const canRemoveNativeElement = (element, parent) => element.parentElement === parent;
const removeNativeElement = (element, parent) => parent.removeChild(element);
//# sourceMappingURL=dom.js.map
