import {
  REF_ATTR,
  ATTR_BLACK_LIST,
  CREATE_EFFECT_TAG,
  UPDATE_EFFECT_TAG,
  DELETE_EFFECT_TAG,
  SKIP_EFFECT_TAG,
  MOVE_MASK,
  FLUSH_MASK,
  SHADOW_MASK,
  detectIsUndefined,
  detectIsBoolean,
  detectIsObject,
  NodeType,
  detectIsTagVirtualNode,
  detectIsTextVirtualNode,
  detectIsPlainVirtualNode,
  getFiberWithElement,
  collectElements,
  walk,
  dummyFn,
  $$scope,
  applyRef,
} from '@dark-engine/core';
import { detectIsSvgElement, detectIsVoidElement } from '../utils';
import { delegateEvent, detectIsEvent, getEventName } from '../events';
import { detectIsPortal } from '../portal';
import {
  INPUT_TAG,
  TEXTAREA_TAG,
  STYLE_ATTR,
  CLASS_ATTR,
  CLASS_NAME_ATTR,
  VALUE_ATTR,
  AS_ATTR,
  EXCLUDE_ATTR_MARK,
} from '../constants';
let moves = [];
let patches = [];
let trackUpdate = null;
const createNativeElementMap = {
  [NodeType.TAG]: vNode => {
    const tagNode = vNode;
    const name = tagNode.name;
    return detectIsSvgElement(name)
      ? document.createElementNS('http://www.w3.org/2000/svg', name)
      : document.createElement(name);
  },
  [NodeType.TEXT]: vNode => {
    return document.createTextNode(vNode.value);
  },
  [NodeType.COMMENT]: vNode => {
    return document.createComment(vNode.value);
  },
};
function createNativeElement(node) {
  return createNativeElementMap[node.type](node);
}
function setObjectStyle(element, style) {
  const keys = Object.keys(style);
  for (const key of keys) {
    element.style.setProperty(key, String(style[key]));
  }
}
function addAttributes(element, node, isHydrateZone) {
  const attrNames = Object.keys(node.attrs);
  const tagElement = element;
  for (let attrName of attrNames) {
    const attrValue = node.attrs[attrName];
    const attribute = performAttribute(tagElement, attrName, attrValue);
    if (attribute === null) {
      continue;
    } else {
      attrName = attribute;
    }
    if (detectIsEvent(attrName)) {
      delegateEvent(tagElement, getEventName(attrName), attrValue);
    } else if (!isHydrateZone && !detectIsUndefined(attrValue) && !ATTR_BLACK_LIST[attrName]) {
      const stop = patchProperties({
        tagName: node.name,
        element: tagElement,
        attrValue,
        attrName,
      });
      !stop && tagElement.setAttribute(attrName, attrValue);
    }
  }
}
function updateAttributes(element, prevNode, nextNode) {
  const attrNames = getAttributeNames(prevNode, nextNode);
  const tagElement = element;
  for (let attrName of attrNames) {
    const prevAttrValue = prevNode.attrs[attrName];
    const nextAttrValue = nextNode.attrs[attrName];
    const attribute = performAttribute(tagElement, attrName, nextAttrValue, prevAttrValue);
    if (attribute === null) {
      continue;
    } else {
      attrName = attribute;
    }
    if (!detectIsUndefined(nextAttrValue)) {
      if (detectIsEvent(attrName)) {
        prevAttrValue !== nextAttrValue && delegateEvent(tagElement, getEventName(attrName), nextAttrValue);
      } else if (!ATTR_BLACK_LIST[attrName] && prevAttrValue !== nextAttrValue) {
        const stop = patchProperties({
          tagName: nextNode.name,
          element: tagElement,
          attrValue: nextAttrValue,
          attrName,
        });
        !stop && tagElement.setAttribute(attrName, nextAttrValue);
      }
    } else {
      tagElement.removeAttribute(attrName);
    }
  }
}
function performAttribute(tagElement, attrName, nextAttrValue, prevAttrValue) {
  if (attrName[0] === EXCLUDE_ATTR_MARK) return null;
  if (attrName === REF_ATTR) {
    applyRef(nextAttrValue, tagElement);
    return null;
  }
  if ((attrName === CLASS_ATTR || attrName === CLASS_NAME_ATTR) && nextAttrValue !== prevAttrValue) {
    toggleAttribute(tagElement, CLASS_ATTR, nextAttrValue);
    return null;
  }
  if (attrName === STYLE_ATTR && nextAttrValue && nextAttrValue !== prevAttrValue && detectIsObject(nextAttrValue)) {
    setObjectStyle(tagElement, nextAttrValue);
    return null;
  }
  if (attrName === AS_ATTR) {
    attrName = attrName.slice(1, AS_ATTR.length);
  }
  return attrName;
}
function toggleAttribute(element, name, value) {
  value ? element.setAttribute(name, value) : element.removeAttribute(name);
}
function getAttributeNames(prevNode, nextNode) {
  const attrNames = new Set();
  const prevAttrs = Object.keys(prevNode.attrs);
  const nextAttrs = Object.keys(nextNode.attrs);
  const size = Math.max(prevAttrs.length, nextAttrs.length);
  for (let i = 0; i < size; i++) {
    attrNames.add(prevAttrs[i] || nextAttrs[i]);
  }
  return attrNames;
}
function patchProperties(options) {
  const { tagName, element, attrName, attrValue } = options;
  const fn = specialCasesMap[tagName];
  let stop = fn ? fn(element, attrName, attrValue) : false;
  if (canSetProperty(element, attrName)) {
    element[attrName] = attrValue;
  }
  if (!stop && detectIsBoolean(attrValue)) {
    stop = !attrName.includes('-');
  }
  return stop;
}
function canSetProperty(element, key) {
  const prototype = Object.getPrototypeOf(element);
  const descriptor = Object.getOwnPropertyDescriptor(prototype, key);
  return Boolean(descriptor?.set);
}
const specialCasesMap = {
  [INPUT_TAG]: (element, attrName, attrValue) => {
    if (attrName === VALUE_ATTR) {
      patches.push(() => {
        detectIsBoolean(attrValue) ? (element.checked = attrValue) : (element.value = String(attrValue));
      });
    }
    return false;
  },
  [TEXTAREA_TAG]: (element, attrName, attrValue) => {
    if (attrName === VALUE_ATTR) {
      element.innerText = String(attrValue);
      return true;
    }
    return false;
  },
};
function commitCreation(fiber) {
  const parentFiber = getFiberWithElement(fiber.parent);
  const parentElement = parentFiber.element;
  const childNodes = parentElement.childNodes;
  const isHydrateZone = $$scope().getIsHydrateZone();
  if (isHydrateZone) {
    const nativeElement = childNodes[fiber.eidx];
    if (
      detectIsTextVirtualNode(fiber.inst) &&
      nativeElement instanceof Text &&
      fiber.inst.value.length !== nativeElement.length
    ) {
      nativeElement.splitText(fiber.inst.value.length);
    }
    fiber.element = nativeElement;
  } else {
    if (!(fiber.mask & SHADOW_MASK)) {
      if (childNodes.length === 0 || fiber.eidx > childNodes.length - 1) {
        !detectIsVoidElement(parentFiber.inst.name) && appendNativeElement(fiber.element, parentElement);
      } else {
        insertNativeElement(fiber.element, parentElement.childNodes[fiber.eidx], parentElement);
      }
    }
  }
  detectIsTagVirtualNode(fiber.inst) && addAttributes(fiber.element, fiber.inst, isHydrateZone);
}
function commitUpdate(fiber) {
  const element = fiber.element;
  const prevInstance = fiber.alt.inst;
  const nextInstance = fiber.inst;
  detectIsPlainVirtualNode(nextInstance)
    ? prevInstance.value !== nextInstance.value && (element.textContent = nextInstance.value)
    : updateAttributes(element, prevInstance, nextInstance);
}
function commitDeletion(fiber) {
  const parentFiber = getFiberWithElement(fiber.parent);
  if (fiber.mask & FLUSH_MASK) {
    parentFiber.element.innerHTML && (parentFiber.element.innerHTML = '');
    return;
  }
  walk(fiber, (fiber, skip) => {
    if (fiber.element) {
      !detectIsPortal(fiber.inst) &&
        canRemoveNativeElement(fiber.element, parentFiber.element) &&
        removeNativeElement(fiber.element, parentFiber.element);
      return skip();
    }
  });
}
function move(fiber) {
  const sourceNodes = collectElements(fiber, x => x.element);
  const sourceNode = sourceNodes[0];
  const parentElement = sourceNode.parentElement;
  const sourceFragment = new DocumentFragment();
  const elementIdx = fiber.eidx;
  let idx = 0;
  const move = () => {
    for (let i = 1; i < sourceNodes.length; i++) {
      removeNativeElement(parentElement.childNodes[elementIdx + 1], parentElement);
    }
    replaceNativeElement(sourceFragment, parentElement.childNodes[elementIdx], parentElement);
  };
  for (const node of sourceNodes) {
    insertNativeElement(document.createComment(`${elementIdx}:${idx}`), node, parentElement);
    appendNativeElement(node, sourceFragment);
    idx++;
  }
  moves.push(move);
}
const commitMap = {
  [CREATE_EFFECT_TAG]: fiber => {
    if (!fiber.element || detectIsPortal(fiber.inst)) return;
    trackUpdate && trackUpdate(fiber.element);
    commitCreation(fiber);
  },
  [UPDATE_EFFECT_TAG]: fiber => {
    fiber.mask & MOVE_MASK && (move(fiber), (fiber.mask &= ~MOVE_MASK));
    if (!fiber.element || detectIsPortal(fiber.inst)) return;
    trackUpdate && trackUpdate(fiber.element);
    commitUpdate(fiber);
  },
  [DELETE_EFFECT_TAG]: commitDeletion,
  [SKIP_EFFECT_TAG]: dummyFn,
};
function finishCommit() {
  moves.forEach(x => x());
  patches.forEach(x => x());
  moves = [];
  patches = [];
}
const commit = fiber => commitMap[fiber.tag](fiber);
const setTrackUpdate = fn => (trackUpdate = fn);
const appendNativeElement = (element, parent) => parent.appendChild(element);
const insertNativeElement = (element, sibling, parent) => parent.insertBefore(element, sibling);
const insertNativeElementByIndex = (element, idx, parent) => parent.insertBefore(element, parent.childNodes[idx]);
const replaceNativeElement = (element, candidate, parent) => parent.replaceChild(element, candidate);
const canRemoveNativeElement = (element, parent) => element.parentElement === parent;
const removeNativeElement = (element, parent) => parent.removeChild(element);
export { createNativeElement, commit, finishCommit, setTrackUpdate, insertNativeElementByIndex };
//# sourceMappingURL=dom.js.map
