import { getTime, workLoop, TaskPriority, detectIsBusy, detectIsFunction, nextTick } from '@dark-engine/core';
import { YIELD_INTERVAL } from '../constants';
class Scheduler {
  queue = {
    [TaskPriority.HIGH]: [],
    [TaskPriority.NORMAL]: [],
    [TaskPriority.LOW]: [],
  };
  deadline = 0;
  task = null;
  scheduledCallback = null;
  isMessageLoopRunning = false;
  channel = null;
  port = null;
  setupPorts() {
    if (process.env.NODE_ENV === 'test') {
      const worker = require('node:worker_threads');
      this.channel = new worker.MessageChannel();
      this.port = this.channel.port2;
      this.channel.port1.on('message', this.performWorkUntilDeadline.bind(this));
      return;
    }
    this.channel = new MessageChannel();
    this.port = this.channel.port2;
    this.channel.port1.onmessage = this.performWorkUntilDeadline.bind(this);
  }
  unrefPorts() {
    this.channel.port1.unref();
    this.channel.port2.unref();
  }
  shouldYield() {
    return getTime() >= this.deadline;
  }
  schedule(callback, options) {
    const {
      priority = TaskPriority.NORMAL,
      forceAsync = false,
      isTransition = false,
      createSign,
      setPendingStatus,
    } = options || {};
    const task = new Task(callback, priority, forceAsync);
    task.setIsTransition(isTransition);
    task.setPendingSetter(setPendingStatus);
    task.setSignCreator(createSign);
    this.put(task);
    this.execute();
  }
  hasPrimaryTask() {
    if (this.task.getIsTransition()) {
      const { high, normal, low } = this.getQueues();
      const hasPrimary = high.length > 0 || normal.length > 0;
      if (hasPrimary) {
        const sign = this.task.createSign();
        const isMatch = Task.detectIsMatchSign(sign, high);
        if (isMatch) {
          this.task.markAsUnnecessary();
        }
        return true;
      }
      const sign = this.task.createSign();
      const hasSame = Task.detectHasSameSign(sign, low);
      if (hasSame) {
        this.task.markAsUnnecessary();
        return true;
      }
    }
    return false;
  }
  cancelTask(fn) {
    if (this.task.getIsUnnecessary()) return this.completeTask(this.task);
    this.task.setTaskRestorer(fn);
    this.deferTransition(this.task);
  }
  completeTask(task) {
    task.pending(false);
  }
  put(task) {
    const queue = this.queue[task.getPriority()];
    queue.push(task);
  }
  pick(queue) {
    if (!queue.length) return false;
    this.task = queue.shift();
    if (this.task.getIsTransition()) {
      const sign = this.task.createSign();
      const hasSign = Task.detectHasSameSign(sign, queue);
      if (hasSign) {
        this.task = null;
        return this.pick(queue);
      }
      if (this.task.canPending()) {
        const task = this.task;
        task.markAsPending();
        this.deferTransition(this.task);
        this.task = null;
        nextTick(() => task.pending(true));
        return true;
      }
    }
    this.task.run();
    this.task.getForceAsync() ? this.requestCallbackAsync(workLoop) : this.requestCallback(workLoop);
    return true;
  }
  execute() {
    const isBusy = detectIsBusy();
    const { high, normal, low } = this.getQueues();
    if (!isBusy && !this.isMessageLoopRunning) {
      this.pick(high) || this.pick(normal) || this.pick(low);
    }
  }
  requestCallbackAsync(callback) {
    this.scheduledCallback = callback;
    if (!this.isMessageLoopRunning) {
      this.isMessageLoopRunning = true;
      this.port.postMessage(null);
    }
  }
  requestCallback(callback) {
    callback(false);
    this.task = null;
    this.execute();
  }
  performWorkUntilDeadline() {
    if (this.scheduledCallback) {
      this.deadline = getTime() + YIELD_INTERVAL;
      const hasMoreWork = this.scheduledCallback(true);
      if (!hasMoreWork) {
        this.completeTask(this.task);
        this.isMessageLoopRunning = false;
        this.scheduledCallback = null;
        this.task = null;
        this.execute();
      } else {
        this.port.postMessage(null);
      }
    } else {
      this.isMessageLoopRunning = false;
    }
  }
  deferTransition(task) {
    const { low } = this.getQueues();
    low.unshift(task);
  }
  getQueues() {
    const high = this.queue[TaskPriority.HIGH];
    const normal = this.queue[TaskPriority.NORMAL];
    const low = this.queue[TaskPriority.LOW];
    return {
      high,
      normal,
      low,
    };
  }
}
class Task {
  id;
  priority;
  forceAsync;
  isTransition;
  isPending;
  isUnnecessary;
  callback;
  taskRestorer = null;
  signCreator;
  pendingSetter = null;
  static nextTaskId = 0;
  constructor(callback, priority, forceAsync) {
    this.id = ++Task.nextTaskId;
    this.callback = callback;
    this.priority = priority;
    this.forceAsync = forceAsync;
  }
  getPriority() {
    return this.priority;
  }
  getForceAsync() {
    return this.forceAsync;
  }
  setIsTransition(value) {
    this.isTransition = value;
  }
  getIsTransition() {
    return this.isTransition;
  }
  run() {
    this.callback(this.taskRestorer);
    this.taskRestorer = null;
  }
  pending(value) {
    this.isTransition && this.pendingSetter && this.pendingSetter(value);
  }
  markAsPending() {
    this.isPending = true;
  }
  canPending() {
    return !this.isPending && detectIsFunction(this.pendingSetter);
  }
  markAsUnnecessary() {
    this.isUnnecessary = true;
  }
  getIsUnnecessary() {
    return this.isUnnecessary;
  }
  setTaskRestorer(fn) {
    this.taskRestorer = fn;
  }
  setSignCreator(fn) {
    this.signCreator = fn;
  }
  createSign() {
    return this.signCreator();
  }
  setPendingSetter(fn) {
    this.pendingSetter = fn;
  }
  static detectHasSameSign(sign, tasks) {
    return tasks.some(x => x.isTransition && x.createSign() === sign);
  }
  static detectIsMatchSign(sign, tasks) {
    return tasks.some(x => x.createSign && x.createSign().length > sign.length);
  }
}
const scheduler = new Scheduler();
if (process.env.NODE_ENV !== 'test') {
  scheduler.setupPorts();
}
export { scheduler };
//# sourceMappingURL=scheduler.js.map
